# bio-conduit
#
# Copyright (C) 2015
#
# author: Natapol Pornputtapong <natapol.por@gmail.com>
#
# Documentation: Natapol Pornputtapong (RDoc'd and embellished by William Webber)
#
module Bio
    module Conduit
        class Step

            attr_reader :name, :info, :resources

            UNITCONV = {
                '' => 1,
                'b' => 1,
                'k' => 1024,
                'm' => 1048576,
                'g' => 1073741824,
                't' => 1099511627776
            }

            def initialize(name, hash, resources, addpath = true)
                @name = name
                @info = hash
                @resources = resources

                if addpath
                    if @info['run'].is_a?(String)
                        @info['run'] = ["mkdir -p <sample_output_path/>#{@name}", "cd <sample_output_path/>#{@name}", @info['run']]
                    elsif @info['run'].is_a?(Array)
                        @info['run'].unshift("mkdir -p <sample_output_path/>#{@name}", "cd <sample_output_path/>#{@name}")
                    end
                end
            end

            def dependence?
                return @info.has_key?('pre')
            end

            def dependency
                if self.dependence?
                    return @info['pre']
                else
                    return nil
                end
            end

            def +(other)
                name = self.name + "_" + other.name
                resources = self.resources.merge(other.resources)
                info = self.info.merge(other.info) do |k, first, second|
                    case k
                    when 'mem', 'cpu', 'nodes'
                        matched1 = /\A(\d+)([kmgtb]?)/.match(first)
                        matched2 = /\A(\d+)([kmgtb]?)/.match(second)
                        if (matched1[1].to_i * Bio::Conduit::Step::UNITCONV[matched1[2]]) > (matched2[1].to_i * Bio::Conduit::Step::UNITCONV[matched2[2]])
                            first
                        elsif (matched1[1].to_i * Bio::Conduit::Step::UNITCONV[matched1[2]]) < (matched2[1].to_i * Bio::Conduit::Step::UNITCONV[matched2[2]])
                            second
                        else
                            first
                        end
                    else
                        if first.is_a?(String) && second.is_a?(String)
                            [first, second]
                        elsif first.is_a?(Array) && second.is_a?(String)
                            first + [second]
                        elsif first.is_a?(String) && second.is_a?(Array)
                            [first] + second
                        else
                            first + second
                        end
                    end
                end
                Bio::Conduit::Step.new(name, info, resources, false)
            end

            def create_jobscript(sample, template)
                total_resources = @resources.merge(sample['resources'])
                total_resources['sample_path'] = sample['path']
                total_resources['sample'] = sample['name']
                total_resources['sample_output_path'] = "#{total_resources['output'].chomp('/')}" + "#{sample.has_key?('group') ? "/#{sample['group']}" : ''}" + "/#{sample['name']}"
                @jobname = sample['name'] + "_" + @name
                @commands = @info['run'].is_a?(Array) ? @info['run'].join("\n") : @info['run'].clone
                subslist = {}
                @commands.scan(/(\<([\w\/]+)\>)/).each do |entry|
                    entry[1].scan(/[^\/]+/).each do |res|
                        entry[1].sub!(/#{res}/, total_resources[res])
                    end
                    @commands.sub!(/#{entry[0]}/, entry[1])
                end
                return ERB.new(template, nil, '-').result(binding)
            end

        end

        class Process

            attr_reader :steps

            def initialize(filename)
                yaml = YAML::load(File.open(filename))
                @steps = {}
                @name = yaml['pipeline']
                @resources = yaml['resources']
                @runninglist = []
                yaml['steps'].each_pair do |name, detail|
                    @resources[name] = "../#{name}"
                    @steps[name] = Bio::Conduit::Step.new(name, detail, @resources)
                end
            end

            def [](name)
                @steps[name]
            end

            def add_running_step(stepname)
                step = self[stepname]
                self.add_running_step(step.dependency) if step.dependence?
                if !@runninglist.include?(stepname)
                    @runninglist.push(stepname)
                end
            end

            def run_steps(steps)

                @steps.keys.each do |name|
                    self.add_running_step(name) if steps == nil || steps.include?(name)
                end

                combinestep = nil

                @runninglist.each do |stepname|
                    if combinestep
                        combinestep = combinestep + self[stepname]
                    else
                        combinestep = self[stepname]
                    end
                end

                return combinestep
            end
        end
    end
end
